/*
 *                            COPYRIGHT
 *
 *  cschem - modular/flexible schematics editor - libcschem (core library)
 *  Copyright (C) 2018 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/cschem
 *    lead developer: email to cschem (at) igor2.repo.hu
 *    mailing list: cschem (at) list.repo.hu (send "subscribe")
 */


#include <stdio.h>
#include <string.h>
#include <librnd/core/safe_fs.h>
#include <librnd/core/plugins.h>
#include <libcschem/config.h>
#include <libcschem/plug_io.h>
#include <libcschem/cnc_loop.h>
#include <libcschem/cnc_line.h>
#include <libcschem/cnc_pen.h>
#include <libcschem/cnc_any_obj.h>
#include <libcschem/rtree.h>
#include <libcschem/message.h>

static csch_plug_io_t eanim;
typedef struct {
	const csch_sheet_t *sheet;
	const char *fn;
	FILE *f;
	const csch_cpen_t *pen;  /* currently active (last selected) pen */
} ctx_t;

static int anim_export_prio(const char *fn, const char *fmt, csch_plug_io_type_t type)
{
	if (type != CSCH_IOTYP_SHEET)
		return 0;
	if (strcmp(fmt, "animator") == 0)
		return 100;
	if (strcmp(fmt, "anim") == 0)
		return 90;
	return 0;
}

static void anim_set_pen(ctx_t *ctx, const csch_cpen_t *pen)
{
	if (ctx->pen == pen)
		return;
	if ((ctx->pen == NULL) || (ctx->pen->color.packed != pen->color.packed))
		fprintf(ctx->f, "color #%02x%02x%02x\n", pen->color.r, pen->color.g, pen->color.b);
	ctx->pen = pen;
}

static void anim_draw_obj(ctx_t *ctx, csch_chdr_t *h)
{
	const csch_cpen_t *pen;


	switch(h->type) {
		case CSCH_CTYPE_GRP:
		case CSCH_CTYPE_GRP_REF:
			/* do not draw anything, no need to recurse either: rtree is global */
			break;
		case CSCH_CTYPE_LINE:
			{
				csch_line_t *l = (csch_line_t *)h;
				csch_coord_t x1 = l->inst.c.p1.x, y1 = l->inst.c.p1.y, x2 = l->inst.c.p2.x, y2 = l->inst.c.p2.y;
				int verts = 10;

				pen = csch_stroke_(h);
				if (pen == NULL) {
					csch_error("omitting object: invalid pen\n");
					return;
				}
				anim_set_pen(ctx, pen);

				switch(ctx->pen->shape) {
					case CSCH_PSHP_ROUND:
						if (ctx->pen->size > 50)
							verts = 20;
						fprintf(ctx->f, "line %ld %ld %ld %ld\n", (long)x1, (long)y1, (long)x2, (long)y2);
						fprintf(ctx->f, "fillcircle %ld %ld %ld %ld\n", (long)x1, (long)y1, (long)ctx->pen->size/2, verts);
						fprintf(ctx->f, "fillcircle %ld %ld %ld %ld\n", (long)x2, (long)y2, (long)ctx->pen->size/2, verts);
						break;
					case CSCH_PSHP_SQUARE:
						abort();
				}
			}
			break;
		case CSCH_CTYPE_TEXT:
			TODO("draw text with vector font");
			break;
	}
}

static void anim_draw_rtree(ctx_t *ctx, const csch_rtree_t *rtr)
{
	csch_rtree_it_t it;
	csch_chdr_t *h;
	for(h = csch_rtree_all_first(&it, rtr); h != NULL; h = csch_rtree_all_next(&it))
		anim_draw_obj(ctx, h);
}

static int anim_export_sheet(const char *fn, const char *fmt, csch_sheet_t *src)
{
	TODO("get hidlib as an arg")
	rnd_hidlib_t *hidlib = NULL;
	FILE *f;
	csch_grpo_iter_t it;
	ctx_t ctx;
	const csch_rtree_box_t *shbb;
	csch_coord_t mx, my;
	int n;

	f = rnd_fopen(hidlib, fn, "w");
	if (f == NULL)
		return -1;

	fprintf(f, "! cschem sheet export\n");

	/* zoom to extents + 5% margin */
	shbb = csch_sheet_bbox(src);
	mx = (shbb->x2 - shbb->x1) / 20;
	my = (shbb->y2 - shbb->y1) / 20;
	fprintf(f, "viewport %ld %ld - %ld %ld\n", (long)(shbb->x1-mx), (long)(shbb->y1-my), (long)(shbb->x2+mx), (long)(shbb->y2+my));

	memset(&ctx, 0, sizeof(ctx));
	ctx.fn = fn;
	ctx.f = f;
	ctx.sheet = src;

	for(n = 0; n < CSCH_DSPLY_max; n++) {
		fprintf(ctx.f, "! display layer: %d: %s\n", n, csch_dsply_name(n));
		anim_draw_rtree(&ctx, &ctx.sheet->dsply[n]);
	}

	fprintf(f, "flush\n");
	fclose(f);
	return 0;
}


int pplg_check_ver_export_animator(int ver_needed) { return 0; }

void pplg_uninit_export_animator(void)
{
	csch_plug_io_unregister(&eanim);
}

int pplg_init_export_animator(void)
{
	RND_API_CHK_VER;

	eanim.name = "export to animator";
	eanim.export_prio = anim_export_prio;
	eanim.export_sheet = anim_export_sheet;
	eanim.ext_export_sheet = ".anim";
	csch_plug_io_register(&eanim);
	return 0;
}

